@@ -52,7 +52,6 @@ function configure_kuryr {
     sudo install -o "$STACK_USER" -m 640 -D "${KURYR_HOME}/etc/kuryr.conf.sample" \
         "$KURYR_CONFIG"
 
-    iniset "$KURYR_CONFIG" kubernetes api_root "$KURYR_K8S_API_URL"
     if [ "$KURYR_K8S_API_CERT" ]; then
         iniset "$KURYR_CONFIG" kubernetes ssl_client_crt_file "$KURYR_K8S_API_CERT"
     fi
@@ -61,6 +60,10 @@ function configure_kuryr {
     fi
     if [ "$KURYR_K8S_API_CACERT" ]; then
         iniset "$KURYR_CONFIG" kubernetes ssl_ca_crt_file "$KURYR_K8S_API_CACERT"
+        iniset "$KURYR_CONFIG" kubernetes ssl_verify_server_crt True
+    fi
+    if [ "$KURYR_MULTI_VIF_DRIVER" ]; then
+        iniset "$KURYR_CONFIG" kubernetes multi_vif_drivers "$KURYR_MULTI_VIF_DRIVER"
     fi
     # REVISIT(ivc): 'use_stderr' is required for current CNI driver. Once a
     # daemon-based CNI driver is implemented, this could be removed.
@@ -70,6 +73,14 @@ function configure_kuryr {
 
     iniset "$KURYR_CONFIG" kubernetes port_debug "$KURYR_PORT_DEBUG"
 
+    iniset "$KURYR_CONFIG" kubernetes pod_subnets_driver "$KURYR_SUBNET_DRIVER"
+    iniset "$KURYR_CONFIG" kubernetes pod_security_groups_driver "$KURYR_SG_DRIVER"
+    iniset "$KURYR_CONFIG" kubernetes service_security_groups_driver "$KURYR_SG_DRIVER"
+    iniset "$KURYR_CONFIG" kubernetes enabled_handlers "$KURYR_ENABLED_HANDLERS"
+
+    # Let Kuryr retry connections to K8s API for 20 minutes.
+    iniset "$KURYR_CONFIG" kubernetes watch_retry_timeout 1200
+
     KURYR_K8S_CONTAINERIZED_DEPLOYMENT=$(trueorfalse False KURYR_K8S_CONTAINERIZED_DEPLOYMENT)
     if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "True" ]; then
         # This works around the issue of being unable to set oslo.privsep mode
@@ -87,9 +98,14 @@ function configure_kuryr {
     fi
 
     if is_service_enabled kuryr-daemon; then
-        iniset "$KURYR_CONFIG" cni_daemon daemon_enabled True
         iniset "$KURYR_CONFIG" oslo_concurrency lock_path "$KURYR_LOCK_DIR"
         create_kuryr_lock_dir
+        if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "False" ]; then
+            iniset "$KURYR_CONFIG" cni_health_server cg_path \
+                "/system.slice/system-devstack.slice/devstack@kuryr-daemon.service"
+        fi
+    else
+        iniset "$KURYR_CONFIG" cni_daemon daemon_enabled False
     fi
 
     create_kuryr_cache_dir
@@ -120,19 +136,34 @@ function configure_kuryr {
 }
 
 function generate_containerized_kuryr_resources {
+    local cni_daemon
+    cni_daemon=$1
+    if [[ KURYR_CONTROLLER_REPLICAS -eq 1 ]]; then
+        KURYR_CONTROLLER_HA="False"
+    else
+        KURYR_CONTROLLER_HA="True"
+    fi
+
     # Containerized deployment will use tokens provided by k8s itself.
     inicomment "$KURYR_CONFIG" kubernetes ssl_client_crt_file
     inicomment "$KURYR_CONFIG" kubernetes ssl_client_key_file
 
+    iniset "$KURYR_CONFIG" kubernetes controller_ha ${KURYR_CONTROLLER_HA}
+    iniset "$KURYR_CONFIG" kubernetes controller_ha_port ${KURYR_CONTROLLER_HA_PORT}
+
+    # NOTE(dulek): In the container the CA bundle will be mounted in a standard
+    # directory, so we need to modify that.
+    iniset "$KURYR_CONFIG" neutron cafile /etc/ssl/certs/kuryr-ca-bundle.crt
     iniset "$KURYR_CONFIG" kubernetes token_file /var/run/secrets/kubernetes.io/serviceaccount/token
     iniset "$KURYR_CONFIG" kubernetes ssl_ca_crt_file /var/run/secrets/kubernetes.io/serviceaccount/ca.crt
 
     # Generate kuryr resources in k8s formats.
     local output_dir="${DATA_DIR}/kuryr-kubernetes"
     generate_kuryr_configmap $output_dir $KURYR_CONFIG $KURYR_CONFIG
+    generate_kuryr_certificates_secret $output_dir $SSL_BUNDLE_FILE
     generate_kuryr_service_account $output_dir
-    generate_controller_deployment $output_dir $KURYR_HEALTH_SERVER_PORT
-    generate_cni_daemon_set $output_dir $CNI_BIN_DIR $CNI_CONF_DIR
+    generate_controller_deployment $output_dir $KURYR_HEALTH_SERVER_PORT $KURYR_CONTROLLER_HA
+    generate_cni_daemon_set $output_dir $KURYR_CNI_HEALTH_SERVER_PORT $cni_daemon $CNI_BIN_DIR $CNI_CONF_DIR
 }
 
 function run_containerized_kuryr_resources {
@@ -140,13 +171,16 @@ function run_containerized_kuryr_resources {
     /usr/local/bin/kubectl create -f \
         "${k8s_data_dir}/config_map.yml" \
         || die $LINENO "Failed to create kuryr-kubernetes ConfigMap."
+    /usr/local/bin/kubectl create -f \
+        "${k8s_data_dir}/certificates_secret.yml" \
+        || die $LINENO "Failed to create kuryr-kubernetes certificates Secret."
     /usr/local/bin/kubectl create -f \
         "${k8s_data_dir}/service_account.yml" \
         || die $LINENO "Failed to create kuryr-kubernetes ServiceAccount."
 
     if is_service_enabled openshift-master; then
         # NOTE(dulek): For OpenShift add privileged SCC to serviceaccount.
-        /usr/local/bin/oadm policy add-scc-to-user privileged -n kube-system -z kuryr-controller
+        /usr/local/bin/oc adm policy add-scc-to-user privileged -n kube-system -z kuryr-controller
     fi
     /usr/local/bin/kubectl create -f \
         "${k8s_data_dir}/controller_deployment.yml" \
@@ -194,6 +228,7 @@ function create_k8s_api_service {
     local service_cidr
     local kubelet_iface_ip
     local lb_name
+    local use_octavia
     local project_id
 
     project_id=$(get_or_create_project \
@@ -208,21 +243,36 @@ function create_k8s_api_service {
 
     k8s_api_clusterip=$(_cidr_range "$service_cidr" | cut -f1)
 
-    create_load_balancer "$lb_name" "$k8s_api_clusterip" \
-        "$project_id" "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET"
-    create_load_balancer_listener default/kubernetes:443 HTTPS 443 "$lb_name" "$project_id"
-    create_load_balancer_pool default/kubernetes:443 HTTPS ROUND_ROBIN \
-        default/kubernetes:443 "$project_id" "$lb_name"
+    create_load_balancer "$lb_name" "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET"\
+            "$project_id" "$k8s_api_clusterip"
+    create_load_balancer_listener default/kubernetes:${KURYR_K8S_API_LB_PORT} HTTPS ${KURYR_K8S_API_LB_PORT} "$lb_name" "$project_id" 3600000
+    create_load_balancer_pool default/kubernetes:${KURYR_K8S_API_LB_PORT} HTTPS ROUND_ROBIN \
+        default/kubernetes:${KURYR_K8S_API_LB_PORT} "$project_id" "$lb_name"
 
     local api_port
     if is_service_enabled openshift-master; then
-        api_port=8443
+        api_port=${OPENSHIFT_API_PORT}
     else
         api_port=6443
     fi
 
-    create_load_balancer_member "$(hostname)" "$kubelet_iface_ip" "$api_port" \
-        default/kubernetes:443 public-subnet "$lb_name" "$project_id"
+    local address
+    KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE=$(trueorfalse True KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE)
+    if [[ "$KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE" == "True" ]]; then
+        address=${kubelet_iface_ip}
+    else
+        address="${HOST_IP}"
+    fi
+
+    use_octavia=$(trueorfalse True KURYR_K8S_LBAAS_USE_OCTAVIA)
+    if [[ "$use_octavia" == "True" && \
+          "$KURYR_K8S_OCTAVIA_MEMBER_MODE" == "L2" ]]; then
+        create_load_balancer_member "$(hostname)" "$address" "$api_port" \
+            default/kubernetes:${KURYR_K8S_API_LB_PORT} $KURYR_NEUTRON_DEFAULT_POD_SUBNET "$lb_name" "$project_id"
+    else
+        create_load_balancer_member "$(hostname)" "$address" "$api_port" \
+            default/kubernetes:${KURYR_K8S_API_LB_PORT} public-subnet "$lb_name" "$project_id"
+    fi
 }
 
 function configure_neutron_defaults {
@@ -232,11 +282,16 @@ function configure_neutron_defaults {
     local service_subnet_id
     local subnetpool_id
     local router
+    local router_id
+    local ext_svc_net_id
+    local ext_svc_subnet_id
 
     # If a subnetpool is not passed, we get the one created in devstack's
     # Neutron module
     subnetpool_id=${KURYR_NEUTRON_DEFAULT_SUBNETPOOL_ID:-${SUBNETPOOL_V4_ID}}
     router=${KURYR_NEUTRON_DEFAULT_ROUTER:-$Q_ROUTER_NAME}
+    router_id="$(openstack router show -c id -f value \
+        "$router")"
 
     project_id=$(get_or_create_project \
         "$KURYR_NEUTRON_DEFAULT_PROJECT" default)
@@ -245,28 +300,31 @@ function configure_neutron_defaults {
                       "$KURYR_NEUTRON_DEFAULT_POD_SUBNET" \
                       "$subnetpool_id" \
                       "$router"
-    pod_subnet_id="$(neutron subnet-show -c id -f value \
+    pod_subnet_id="$(openstack subnet show -c id -f value \
         "${KURYR_NEUTRON_DEFAULT_POD_SUBNET}")"
 
+    local use_octavia
+    use_octavia=$(trueorfalse True KURYR_K8S_LBAAS_USE_OCTAVIA)
     create_k8s_subnet "$project_id" \
                       "$KURYR_NEUTRON_DEFAULT_SERVICE_NET" \
                       "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET" \
                       "$subnetpool_id" \
-                      "$router"
-    service_subnet_id="$(neutron subnet-show -c id -f value \
+                      "$router" \
+                      "$use_octavia"
+    service_subnet_id="$(openstack subnet show -c id -f value \
         "${KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET}")"
 
-    sg_ids=$(echo $(neutron security-group-list \
-        --project-id "$project_id" -c id -f value) | tr ' ' ',')
+    if [ "$KURYR_SG_DRIVER" != "namespace" ]; then
+        sg_ids=$(echo $(openstack security group list \
+            --project "$project_id" -c ID -f value) | tr ' ' ',')
+    fi
 
-    ext_svc_net_id="$(neutron net-show -c id -f value \
+    ext_svc_net_id="$(openstack network show -c id -f value \
         "${KURYR_NEUTRON_DEFAULT_EXT_SVC_NET}")"
 
-    ext_svc_subnet_id="$(neutron subnet-show -c id -f value \
+    ext_svc_subnet_id="$(openstack subnet show -c id -f value \
         "${KURYR_NEUTRON_DEFAULT_EXT_SVC_SUBNET}")"
 
-    local use_octavia
-    use_octavia=$(trueorfalse True KURYR_K8S_LBAAS_USE_OCTAVIA)
     if [[ "$use_octavia" == "True" && \
           "$KURYR_K8S_OCTAVIA_MEMBER_MODE" == "L3" ]]; then
         # In order for the pods to allow service traffic under Octavia L3 mode,
@@ -285,7 +343,11 @@ function configure_neutron_defaults {
             --description "k8s service subnet allowed" \
             --remote-ip "$service_cidr" --ethertype IPv4 --protocol tcp \
             "$service_pod_access_sg_id"
-        sg_ids+=",${service_pod_access_sg_id}"
+        if [ -n "$sg_ids" ]; then
+            sg_ids+=",${service_pod_access_sg_id}"
+        else
+            sg_ids="${service_pod_access_sg_id}"
+        fi
     elif [[ "$use_octavia" == "True" && \
             "$KURYR_K8S_OCTAVIA_MEMBER_MODE" == "L2" ]]; then
         # In case the member connectivity is L2, Octavia by default uses the
@@ -308,13 +370,72 @@ function configure_neutron_defaults {
             --description "k8s pod subnet allowed from k8s-pod-subnet" \
             --remote-ip "$pod_cidr" --ethertype IPv4 --protocol tcp \
             "$octavia_pod_access_sg_id"
-        sg_ids+=",${octavia_pod_access_sg_id}"
+        if [ -n "$sg_ids" ]; then
+            sg_ids+=",${octavia_pod_access_sg_id}"
+        else
+            sg_ids="${octavia_pod_access_sg_id}"
+        fi
     fi
 
+    KURYR_K8S_CONTAINERIZED_DEPLOYMENT=$(trueorfalse False KURYR_K8S_CONTAINERIZED_DEPLOYMENT)
+    if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "False" ]; then
+        local service_cidr
+        local k8s_api_clusterip
+        service_cidr=$(openstack --os-cloud devstack-admin \
+                                 --os-region "$REGION_NAME" \
+                                 subnet show "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET" \
+                                 -c cidr -f value)
+        k8s_api_clusterip=$(_cidr_range "$service_cidr" | cut -f1)
+        # NOTE(dulek): KURYR_K8S_API_LB_URL will be a global to be used by next
+        #              deployment phases.
+        KURYR_K8S_API_LB_URL="https://${k8s_api_clusterip}:${KURYR_K8S_API_LB_PORT}"
+        iniset "$KURYR_CONFIG" kubernetes api_root ${KURYR_K8S_API_LB_URL}
+    else
+        iniset "$KURYR_CONFIG" kubernetes api_root '""'
+    fi
     iniset "$KURYR_CONFIG" neutron_defaults project "$project_id"
     iniset "$KURYR_CONFIG" neutron_defaults pod_subnet "$pod_subnet_id"
     iniset "$KURYR_CONFIG" neutron_defaults pod_security_groups "$sg_ids"
     iniset "$KURYR_CONFIG" neutron_defaults service_subnet "$service_subnet_id"
+    if [ "$KURYR_SUBNET_DRIVER" == "namespace" ]; then
+        iniset "$KURYR_CONFIG" namespace_subnet pod_subnet_pool "$subnetpool_id"
+        iniset "$KURYR_CONFIG" namespace_subnet pod_router "$router_id"
+    fi
+    if [ "$KURYR_SG_DRIVER" == "namespace" ]; then
+        local allow_namespace_sg_id
+        local allow_default_sg_id
+        allow_namespace_sg_id=$(openstack --os-cloud devstack-admin \
+            --os-region "$REGION_NAME" \
+            security group create --project "$project_id" \
+            allow_from_namespace -f value -c id)
+        allow_default_sg_id=$(openstack --os-cloud devstack-admin \
+            --os-region "$REGION_NAME" \
+            security group create --project "$project_id" \
+            allow_from_default -f value -c id)
+        openstack --os-cloud devstack-admin --os-region "$REGION_NAME" \
+            security group rule create --project "$project_id" \
+            --description "allow traffic from default namespace" \
+            --remote-group "$allow_namespace_sg_id" --ethertype IPv4 --protocol tcp \
+            "$allow_default_sg_id"
+        openstack --os-cloud devstack-admin --os-region "$REGION_NAME" \
+            security group rule create --project "$project_id" \
+            --description "allow traffic from default namespace" \
+            --remote-group "$allow_namespace_sg_id" --ethertype IPv4 --protocol icmp \
+            "$allow_default_sg_id"
+        openstack --os-cloud devstack-admin --os-region "$REGION_NAME" \
+            security group rule create --project "$project_id" \
+            --description "allow traffic from namespaces at default namespace" \
+            --remote-group "$allow_default_sg_id" --ethertype IPv4 --protocol tcp \
+            "$allow_namespace_sg_id"
+        openstack --os-cloud devstack-admin --os-region "$REGION_NAME" \
+            security group rule create --project "$project_id" \
+            --description "allow traffic from namespaces at default namespace" \
+            --remote-group "$allow_default_sg_id" --ethertype IPv4 --protocol icmp \
+            "$allow_namespace_sg_id"
+
+        iniset "$KURYR_CONFIG" namespace_sg sg_allow_from_namespaces "$allow_namespace_sg_id"
+        iniset "$KURYR_CONFIG" namespace_sg sg_allow_from_default "$allow_default_sg_id"
+    fi
     if [ -n "$OVS_BRIDGE" ]; then
         iniset "$KURYR_CONFIG" neutron_defaults ovs_bridge "$OVS_BRIDGE"
     fi
@@ -340,7 +461,7 @@ function configure_k8s_pod_sg_rules {
                       --os-region "$REGION_NAME" \
                       security group list \
                       --project "$project_id" -c ID -c Name -f value | \
-                      awk '/default/ {print $1}')
+                      awk '{if ($2=="default") print $1}')
     create_k8s_icmp_sg_rules "$sg_id" ingress
 }
 
@@ -385,6 +506,11 @@ function prepare_kubernetes_files {
     sudo bash -c "echo '$(create_token),kubelet,kubelet' >> ${KURYR_HYPERKUBE_DATA_DIR}/known_tokens.csv"
     sudo bash -c "echo '$(create_token),kube_proxy,kube_proxy' >> ${KURYR_HYPERKUBE_DATA_DIR}/known_tokens.csv"
 
+    # Copy certs for Kuryr services to use
+    sudo install -m 644 "${KURYR_HYPERKUBE_DATA_DIR}/kubecfg.crt" "${KURYR_HYPERKUBE_DATA_DIR}/kuryr.crt"
+    sudo install -m 644 "${KURYR_HYPERKUBE_DATA_DIR}/kubecfg.key" "${KURYR_HYPERKUBE_DATA_DIR}/kuryr.key"
+    sudo install -m 644 "${KURYR_HYPERKUBE_DATA_DIR}/ca.crt" "${KURYR_HYPERKUBE_DATA_DIR}/kuryr-ca.crt"
+
     # FIXME(ivc): replace 'sleep' with a strict check (e.g. wait_for_files)
     # 'kubernetes-api' fails if started before files are generated.
     # this is a workaround to prevent races.
@@ -396,21 +522,17 @@ function wait_for {
     local url
     local cacert_path
     local flags
-    local timeout=$KURYR_WAIT_TIMEOUT
     name="$1"
     url="$2"
     cacert_path=${3:-}
+    timeout=${4:-$KURYR_WAIT_TIMEOUT}
 
     echo -n "Waiting for $name to respond"
 
-    if [ $# == 3 ]; then
-        extra_flags="--cacert ${cacert_path}"
-    else
-        extra_flags=""
-    fi
+    extra_flags=${cacert_path:+"--cacert ${cacert_path}"}
 
     local start_time=$(date +%s)
-    until curl -o /dev/null -sIf $extra_flags "$url"; do
+    until curl -o /dev/null -s $extra_flags "$url"; do
         echo -n "."
         local curr_time=$(date +%s)
         local time_diff=$(($curr_time - $start_time))
@@ -421,15 +543,21 @@ function wait_for {
 }
 
 function run_k8s_api {
+    local service_cidr
     local cluster_ip_range
 
     # Runs Hyperkube's Kubernetes API Server
     wait_for "etcd" "${KURYR_ETCD_ADVERTISE_CLIENT_URL}/v2/machines"
 
-    cluster_ip_range=$(openstack --os-cloud devstack-admin \
+    service_cidr=$(openstack --os-cloud devstack-admin \
                          --os-region "$REGION_NAME" \
                          subnet show "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET" \
                          -c cidr -f value)
+    if is_service_enabled octavia; then
+        cluster_ip_range=$(split_subnet "$service_cidr" | cut -f1)
+    else
+        cluster_ip_range="$service_cidr"
+    fi
 
     run_container kubernetes-api \
         --net host \
@@ -539,12 +667,17 @@ function run_k8s_kubelet {
     # saves us from the arduous task of setting up mounts to the official image
     # adding Python and all our CNI/binding dependencies.
     local command
+    local minor_version
+    local cgroup_driver
+
+    cgroup_driver="$(docker info|awk '/Cgroup/ {print $NF}')"
 
     sudo mkdir -p "${KURYR_HYPERKUBE_DATA_DIR}/"{kubelet,kubelet.cert}
     command="$KURYR_HYPERKUBE_BINARY kubelet\
         --kubeconfig=${HOME}/.kube/config --require-kubeconfig \
         --allow-privileged=true \
         --v=2 \
+        --cgroup-driver=$cgroup_driver \
         --address=0.0.0.0 \
         --enable-server \
         --network-plugin=cni \
@@ -553,8 +686,9 @@ function run_k8s_kubelet {
         --cert-dir=${KURYR_HYPERKUBE_DATA_DIR}/kubelet.cert \
         --root-dir=${KURYR_HYPERKUBE_DATA_DIR}/kubelet"
 
-    # Kubernetes 1.8 requires additional option to work in the gate.
-    if [[ ${KURYR_HYPERKUBE_VERSION} == v1.8* ]]; then
+    # Kubernetes 1.8+ requires additional option to work in the gate.
+    minor_version=${KURYR_HYPERKUBE_VERSION:3:1}
+    if [ ${minor_version} -gt 7 ]; then
         command="$command --fail-swap-on=false"
     fi
 
@@ -573,13 +707,14 @@ function run_k8s_kubelet {
 
 function run_kuryr_kubernetes {
     local python_bin=$(which python)
-
     if is_service_enabled openshift-master; then
-        wait_for "OpenShift API Server" "$OPENSHIFT_API_URL" \
-            "${OPENSHIFT_DATA_DIR}/ca.crt"
+        wait_for "OpenShift API Server" "$KURYR_K8S_API_LB_URL" \
+            "${OPENSHIFT_DATA_DIR}/ca.crt" 1200
     else
-        wait_for "Kubernetes API Server" "$KURYR_K8S_API_URL"
+        wait_for "Kubernetes API Server" "$KURYR_K8S_API_LB_URL" \
+            "${KURYR_HYPERKUBE_DATA_DIR}/kuryr-ca.crt" 1200
     fi
+
     run_process kuryr-kubernetes \
         "$python_bin ${KURYR_HOME}/scripts/run_server.py  \
             --config-file $KURYR_CONFIG"
@@ -591,6 +726,59 @@ function run_kuryr_daemon {
     run_process kuryr-daemon "$daemon_bin --config-file $KURYR_CONFIG" root root
 }
 
+function create_ingress_l7_router {
+
+    local lb_port_id
+    local lb_name
+    local project_id
+    local max_timeout
+    local lb_vip
+    local fake_svc_name
+    local l7_router_fip
+    local project_id
+    local lb_uuid
+
+    lb_name=${KURYR_L7_ROUTER_NAME}
+    max_timeout=600
+    project_id=$(get_or_create_project \
+        "$KURYR_NEUTRON_DEFAULT_PROJECT" default)
+
+    create_load_balancer "$lb_name" "$KURYR_NEUTRON_DEFAULT_SERVICE_SUBNET" "$project_id"
+
+    wait_for_lb $lb_name $max_timeout
+
+    lb_port_id="$(get_loadbalancer_attribute "$lb_name" "vip_port_id")"
+
+    #allocate FIP and bind it to lb vip
+    l7_router_fip=$(openstack --os-cloud devstack-admin \
+           --os-region "$REGION_NAME" \
+           floating ip create --project "$project_id" \
+            --subnet "${ext_svc_subnet_id}"  "${ext_svc_net_id}" \
+            -f value -c floating_ip_address)
+
+    openstack  --os-cloud devstack-admin \
+            --os-region "$REGION_NAME" \
+            floating ip set --port "$lb_port_id" "$l7_router_fip"
+
+    lb_uuid="$(get_loadbalancer_attribute "$lb_name" "id")"
+    iniset "$KURYR_CONFIG" ingress l7_router_uuid "$lb_uuid"
+
+    #in case tempest enabled, update router's FIP in tempest.conf
+    if is_service_enabled tempest; then
+       iniset $TEMPEST_CONFIG kuryr_kubernetes ocp_router_fip "$l7_router_fip"
+    fi
+
+    if is_service_enabled octavia; then
+        echo -n "Octavia: no need to create fake k8s service for Ingress."
+    else
+        # keep fake an endpoint less k8s service to keep Kubernetes API server
+        # from allocating ingress LB vip
+        fake_svc_name='kuryr-svc-ingress'
+        echo -n "LBaaS: create fake k8s service: $fake_svc_name for Ingress."
+        lb_vip="$(get_loadbalancer_attribute "$lb_name" "vip_address")"
+        create_k8s_fake_service $fake_svc_name $lb_vip
+    fi
+}
 
 source $DEST/kuryr-kubernetes/devstack/lib/kuryr_kubernetes
 
@@ -605,9 +793,7 @@ if [[ "$1" == "stack" && "$2" == "install" ]]; then
     fi
 
 elif [[ "$1" == "stack" && "$2" == "post-config" ]]; then
-    if is_service_enabled kuryr-kubernetes; then
-        create_kuryr_account
-    fi
+    create_kuryr_account
     configure_kuryr
 fi
 
@@ -653,9 +839,17 @@ if [[ "$1" == "stack" && "$2" == "extra" ]]; then
     if is_service_enabled openshift-node; then
         prepare_kubelet
         run_openshift_node
+        if is_service_enabled openshift-dns; then
+            FIRST_NAMESERVER=$(grep nameserver /etc/resolv.conf | awk '{print $2; exit}')
+            openshift_node_set_dns_config "${OPENSHIFT_DATA_DIR}/node/node-config.yaml" \
+                "$FIRST_NAMESERVER"
+            run_openshift_dnsmasq "$FIRST_NAMESERVER"
+            run_openshift_dns
+       fi
+
         KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE=$(trueorfalse True KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE)
         if [[ "$KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE" == "True" ]]; then
-            ovs_bind_for_kubelet "$KURYR_NEUTRON_DEFAULT_PROJECT"
+            ovs_bind_for_kubelet "$KURYR_NEUTRON_DEFAULT_PROJECT" ${OPENSHIFT_API_PORT}
         fi
     fi
 
@@ -679,12 +873,6 @@ if [[ "$1" == "stack" && "$2" == "extra" ]]; then
         run_k8s_scheduler
     fi
 
-    KURYR_K8S_CONTAINERIZED_DEPLOYMENT=$(trueorfalse False KURYR_K8S_CONTAINERIZED_DEPLOYMENT)
-    if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "False" ]; then
-        # If running in containerized mode, we'll run the daemon as DaemonSet.
-        run_kuryr_daemon
-    fi
-
     if is_service_enabled kubelet; then
         prepare_kubelet
         extract_hyperkube
@@ -692,7 +880,7 @@ if [[ "$1" == "stack" && "$2" == "extra" ]]; then
         run_k8s_kubelet
         KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE=$(trueorfalse True KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE)
         if [[ "$KURYR_CONFIGURE_BAREMETAL_KUBELET_IFACE" == "True" ]]; then
-            ovs_bind_for_kubelet "$KURYR_NEUTRON_DEFAULT_PROJECT"
+            ovs_bind_for_kubelet "$KURYR_NEUTRON_DEFAULT_PROJECT" 6443
         fi
     fi
 
@@ -701,16 +889,20 @@ if [[ "$1" == "stack" && "$2" == "extra" ]]; then
         configure_k8s_pod_sg_rules
     fi
 
+    KURYR_K8S_CONTAINERIZED_DEPLOYMENT=$(trueorfalse False KURYR_K8S_CONTAINERIZED_DEPLOYMENT)
     if is_service_enabled kuryr-kubernetes; then
-        if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "False" ]; then
-            run_kuryr_kubernetes
-        else
+        /usr/local/bin/kubectl apply -f ${KURYR_HOME}/kubernetes_crds/kuryrnet.yaml
+        if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "True" ]; then
             if is_service_enabled kuryr-daemon; then
                 build_kuryr_containers $CNI_BIN_DIR $CNI_CONF_DIR True
+                generate_containerized_kuryr_resources True
             else
                 build_kuryr_containers $CNI_BIN_DIR $CNI_CONF_DIR False
+                generate_containerized_kuryr_resources False
             fi
-            generate_containerized_kuryr_resources
+        fi
+        if [ "$KURYR_MULTI_VIF_DRIVER" == "npwg_multiple_interfaces" ]; then
+            /usr/local/bin/kubectl apply -f ${KURYR_HOME}/kubernetes_crds/network_attachment_definition_crd.yaml
         fi
     fi
 
@@ -718,18 +910,50 @@ elif [[ "$1" == "stack" && "$2" == "test-config" ]]; then
     if is_service_enabled kuryr-kubernetes; then
         # NOTE(dulek): This is so late, because Devstack's Octavia is unable
         #              to create loadbalancers until test-config phase.
-        create_k8s_router_fake_service
+        use_octavia=$(trueorfalse True KURYR_K8S_LBAAS_USE_OCTAVIA)
+        if [[ "$use_octavia" == "False" ]]; then
+            create_k8s_router_fake_service
+        fi
         create_k8s_api_service
+        #create Ingress L7 router if required
+        enable_ingress=$(trueorfalse False KURYR_ENABLE_INGRESS)
+
+        if [ "$enable_ingress" == "True" ]; then
+            create_ingress_l7_router
+        fi
 
         # FIXME(dulek): This is a very late phase to start Kuryr services.
         #               We're doing it here because we need K8s API LB to be
-        #               created in order to run kuryr-cni container. Thing is
+        #               created in order to run kuryr services. Thing is
         #               Octavia is unable to create LB until test-config phase.
         #               We can revisit this once Octavia's DevStack plugin will
         #               get improved.
         if [ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "True" ]; then
             run_containerized_kuryr_resources
+        else
+            run_kuryr_kubernetes
+            run_kuryr_daemon
         fi
+
+        # Needs kuryr to be running
+        if is_service_enabled openshift-dns; then
+            configure_and_run_registry
+        fi
+    fi
+    if is_service_enabled tempest && [[ "$KURYR_USE_PORT_POOLS" == "True" ]]; then
+        iniset $TEMPEST_CONFIG kuryr_kubernetes port_pool_enabled True
+    fi
+    if is_service_enabled tempest && [[ "$KURYR_K8S_CONTAINERIZED_DEPLOYMENT" == "True" ]]; then
+        iniset $TEMPEST_CONFIG kuryr_kubernetes containerized True
+    fi
+    if is_service_enabled tempest && [[ "$KURYR_SUBNET_DRIVER" == "namespace" ]]; then
+        iniset $TEMPEST_CONFIG kuryr_kubernetes namespace_enabled True
+    fi
+    if is_service_enabled tempest && [[ "$KURYR_K8S_SERIAL_TESTS" == "True" ]]; then
+        iniset $TEMPEST_CONFIG kuryr_kubernetes run_tests_serial True
+    fi
+    if is_service_enabled tempest && [[ "$KURYR_MULTI_VIF_DRIVER" == "npwg_multiple_interfaces" ]]; then
+        iniset $TEMPEST_CONFIG kuryr_kubernetes npwg_multi_vif_enabled True
     fi
 fi
 
@@ -765,10 +989,20 @@ if [[ "$1" == "unstack" ]]; then
     fi
     if is_service_enabled openshift-node; then
         stop_process openshift-node
+        if is_service_enabled openshift-dns; then
+            reinstate_old_dns_config
+            stop_process openshift-dns
+            stop_process openshift-dnsmasq
+        fi
+        # NOTE(dulek): We need to clean up the configuration as well, otherwise
+        # when doing stack.sh again, openshift-node will use old certificates.
+        sudo rm -rf ${OPENSHIFT_DATA_DIR}
     fi
     if is_service_enabled legacy_etcd; then
         stop_container etcd
     fi
+
+    cleanup_kuryr_devstack_iptables
 fi
 
 if [[ "$1" == "clean" ]]; then
