@@ -13,6 +13,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import os
+
 from kuryr.lib import utils
 
 from kuryr_kubernetes import config
@@ -52,6 +54,7 @@ def setup_loadbalancer_client():
     if any(ext['alias'] == 'lbaasv2' for
            ext in neutron_client.list_extensions()['extensions']):
         _clients[_LB_CLIENT] = neutron_client
+        neutron_client.cascading_capable = False
     else:
         # Since Octavia is lbaasv2 API compatible (A superset of it) we'll just
         # wire an extra neutron client instance to point to it
@@ -63,8 +66,16 @@ def setup_loadbalancer_client():
             service_type='load-balancer')
         lbaas_client.httpclient = octo_httpclient
         _clients[_LB_CLIENT] = lbaas_client
+        lbaas_client.cascading_capable = True
 
 
 def setup_kubernetes_client():
-    _clients[_KUBERNETES_CLIENT] = k8s_client.K8sClient(
-        config.CONF.kubernetes.api_root)
+    if config.CONF.kubernetes.api_root:
+        api_root = config.CONF.kubernetes.api_root
+    else:
+        # NOTE(dulek): This is for containerized deployments, i.e. running in
+        #              K8s Pods.
+        host = os.environ['KUBERNETES_SERVICE_HOST']
+        port = os.environ['KUBERNETES_SERVICE_PORT_HTTPS']
+        api_root = "https://%s:%s" % (host, port)
+    _clients[_KUBERNETES_CLIENT] = k8s_client.K8sClient(api_root)
