@@ -23,6 +23,7 @@ from kuryr_kubernetes import clients
 from kuryr_kubernetes import config
 from kuryr_kubernetes import constants
 from kuryr_kubernetes.controller.drivers import nested_vif
+from kuryr_kubernetes.controller.drivers import utils
 from kuryr_kubernetes import exceptions as k_exc
 from kuryr_kubernetes import os_vif_util as ovu
 
@@ -106,7 +107,7 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
             vifs.append(vif)
         return vifs
 
-    def release_vif(self, pod, vif, project_id=None, security_groups=None):
+    def release_vif(self, pod, vif):
         neutron = clients.get_neutron_client()
         parent_port = self._get_parent_port(neutron, pod)
         trunk_id = self._get_trunk_id(parent_port)
@@ -121,7 +122,7 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
     def _get_port_request(self, pod, project_id, subnets, security_groups,
                           unbound=False):
         port_req_body = {'project_id': project_id,
-                         'network_id': self._get_network_id(subnets),
+                         'network_id': utils.get_network_id(subnets),
                          'fixed_ips': ovu.osvif_to_neutron_fixed_ips(subnets),
                          'device_owner': kl_const.DEVICE_OWNER,
                          'admin_state_up': True}
@@ -131,7 +132,7 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
             if unbound:
                 port_req_body['name'] = constants.KURYR_PORT_NAME
             else:
-                port_req_body['name'] = self._get_port_name(pod)
+                port_req_body['name'] = utils.get_port_name(pod)
 
         if security_groups:
             port_req_body['security_groups'] = security_groups
@@ -212,17 +213,22 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
                 raise ex
             return vlan_id
 
-    def _remove_subport(self, neutron, trunk_id, subport_id):
-        subport_id = [{'port_id': subport_id}]
+    def _remove_subports(self, neutron, trunk_id, subports_id):
+        subports_body = []
+        for subport_id in subports_id:
+            subports_body.append({'port_id': subport_id})
         try:
             neutron.trunk_remove_subports(trunk_id,
-                                          {'sub_ports': subport_id})
+                                          {'sub_ports': subports_body})
         except n_exc.NeutronClientException as ex:
             LOG.error(
                 "Error happened during subport removal from "
                 "trunk, %s", trunk_id)
             raise ex
 
+    def _remove_subport(self, neutron, trunk_id, subport_id):
+        self._remove_subports(neutron, trunk_id, [subport_id])
+
     def _get_vlan_id(self, trunk_id):
         vlan_ids = self._get_in_use_vlan_ids_set(trunk_id)
         return seg_driver.allocate_segmentation_id(vlan_ids)
