@@ -22,6 +22,8 @@ import time
 from kuryr.lib._i18n import _
 from kuryr.lib import constants as kl_const
 from neutronclient.common import exceptions as n_exc
+from oslo_cache import core as cache
+from oslo_concurrency import lockutils
 from oslo_config import cfg as oslo_cfg
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
@@ -30,10 +32,11 @@ from kuryr_kubernetes import clients
 from kuryr_kubernetes import config
 from kuryr_kubernetes import constants
 from kuryr_kubernetes.controller.drivers import base
-from kuryr_kubernetes.controller.drivers import default_subnet
+from kuryr_kubernetes.controller.drivers import utils as c_utils
 from kuryr_kubernetes.controller.managers import pool
 from kuryr_kubernetes import exceptions
 from kuryr_kubernetes import os_vif_util as ovu
+from kuryr_kubernetes import utils
 
 LOG = logging.getLogger(__name__)
 
@@ -53,10 +56,31 @@ vif_pool_driver_opts = [
                     help=_("Minimun interval (in seconds) "
                            "between pool updates"),
                     default=20),
+    oslo_cfg.DictOpt('pools_vif_drivers',
+                     help=_("Dict with the pool driver and pod driver to be "
+                            "used. If not set, it will take them from the "
+                            "kubernetes driver options for pool and pod "
+                            "drivers respectively"),
+                     default={}),
 ]
 
 oslo_cfg.CONF.register_opts(vif_pool_driver_opts, "vif_pool")
 
+node_vif_driver_caching_opts = [
+    oslo_cfg.BoolOpt('caching', default=True),
+    oslo_cfg.IntOpt('cache_time', default=3600),
+]
+
+oslo_cfg.CONF.register_opts(node_vif_driver_caching_opts,
+                            "node_driver_caching")
+
+cache.configure(oslo_cfg.CONF)
+node_driver_cache_region = cache.create_region()
+MEMOIZE = cache.get_memoization_decorator(
+    oslo_cfg.CONF, node_driver_cache_region, "node_driver_caching")
+
+cache.configure_cache_region(oslo_cfg.CONF, node_driver_cache_region)
+
 
 class NoopVIFPool(base.VIFPoolDriver):
     """No pool VIFs for Kubernetes Pods"""
@@ -69,7 +93,7 @@ class NoopVIFPool(base.VIFPoolDriver):
                                          security_groups)
 
     def release_vif(self, pod, vif, *argv):
-        self._drv_vif.release_vif(pod, vif, *argv)
+        self._drv_vif.release_vif(pod, vif)
 
     def activate_vif(self, pod, vif):
         self._drv_vif.activate_vif(pod, vif)
@@ -102,12 +126,14 @@ class BaseVIFPool(base.VIFPoolDriver):
     - ports_pool_update_frequency: interval in seconds between ports pool
     updates, both for populating pools as well as for recycling ports.
     """
-    _available_ports_pools = collections.defaultdict(collections.deque)
-    _existing_vifs = collections.defaultdict(collections.defaultdict)
-    _recyclable_ports = collections.defaultdict(collections.defaultdict)
-    _last_update = collections.defaultdict(collections.defaultdict)
 
     def __init__(self):
+        self._available_ports_pools = collections.defaultdict(
+            collections.deque)
+        self._existing_vifs = collections.defaultdict(collections.defaultdict)
+        self._recyclable_ports = collections.defaultdict(
+            collections.defaultdict)
+        self._last_update = collections.defaultdict(collections.defaultdict)
         # Note(ltomasbo) Execute the port recycling periodic actions in a
         # background thread
         eventlet.spawn(self._return_ports_to_pool)
@@ -135,6 +161,9 @@ class BaseVIFPool(base.VIFPoolDriver):
                     net_id)
         return pool_key
 
+    def _get_pool_key_net(self, pool_key):
+        return pool_key[3]
+
     def request_vif(self, pod, project_id, subnets, security_groups):
         try:
             host_addr = self._get_host_addr(pod)
@@ -209,12 +238,13 @@ class BaseVIFPool(base.VIFPoolDriver):
             try:
                 annotations = jsonutils.loads(pod['metadata']['annotations'][
                     constants.K8S_ANNOTATION_VIF])
+                pod_state = utils.extract_pod_annotation(annotations)
             except KeyError:
                 LOG.debug("Skipping pod without kuryr VIF annotation: %s",
                           pod)
             else:
-                in_use_ports.append(
-                    annotations['versioned_object.data']['id'])
+                for vif in pod_state.vifs.values():
+                    in_use_ports.append(vif.id)
         return in_use_ports
 
     def list_pools(self):
@@ -223,6 +253,9 @@ class BaseVIFPool(base.VIFPoolDriver):
     def show_pool(self, pool_key):
         return self._available_ports_pools.get(pool_key)
 
+    def delete_network_pools(self, net_id):
+        raise NotImplementedError()
+
     def _create_healthcheck_file(self):
         # Note(ltomasbo): Create a health check file when the pre-created
         # ports are loaded into their corresponding pools. This file is used
@@ -253,7 +286,7 @@ class NeutronVIFPool(BaseVIFPool):
                 port_id,
                 {
                     "port": {
-                        'name': pod['metadata']['name'],
+                        'name': c_utils.get_port_name(pod),
                         'device_id': pod['metadata']['uid']
                     }
                 })
@@ -274,54 +307,61 @@ class NeutronVIFPool(BaseVIFPool):
         If a maximun number of port per pool is set, the port will be
         deleted if the maximun has been already reached.
         """
-        neutron = clients.get_neutron_client()
         while True:
-            sg_current = {}
-            if not config.CONF.kubernetes.port_debug:
-                kuryr_ports = self._get_ports_by_attrs(
-                    device_owner=kl_const.DEVICE_OWNER)
-                for port in kuryr_ports:
-                    if port['id'] in self._recyclable_ports.keys():
-                        sg_current[port['id']] = port['security_groups']
-
-            for port_id, pool_key in self._recyclable_ports.copy().items():
-                if (not oslo_cfg.CONF.vif_pool.ports_pool_max or
-                    self._get_pool_size(pool_key) <
-                        oslo_cfg.CONF.vif_pool.ports_pool_max):
-                    port_name = (constants.KURYR_PORT_NAME
-                                 if config.CONF.kubernetes.port_debug
-                                 else '')
-                    if (config.CONF.kubernetes.port_debug or
-                            list(pool_key[2]) != sg_current.get(port_id)):
-                        try:
-                            neutron.update_port(
-                                port_id,
-                                {
-                                    "port": {
-                                        'name': port_name,
-                                        'device_id': '',
-                                        'security_groups': list(pool_key[2])
-                                    }
-                                })
-                        except n_exc.NeutronClientException:
-                            LOG.warning("Error preparing port %s to be "
-                                        "reused, put back on the cleanable "
-                                        "pool.", port_id)
-                            continue
-                    self._available_ports_pools.setdefault(
-                        pool_key, []).append(port_id)
-                else:
-                    try:
-                        del self._existing_vifs[port_id]
-                        neutron.delete_port(port_id)
-                    except n_exc.PortNotFoundClient:
-                        LOG.debug('Unable to release port %s as it no longer '
-                                  'exists.', port_id)
-                    except KeyError:
-                        LOG.debug('Port %s is not in the ports list.', port_id)
-                del self._recyclable_ports[port_id]
+            self._trigger_return_to_pool()
             eventlet.sleep(oslo_cfg.CONF.vif_pool.ports_pool_update_frequency)
 
+    @lockutils.synchronized('return_to_pool_baremetal')
+    def _trigger_return_to_pool(self):
+        neutron = clients.get_neutron_client()
+        sg_current = {}
+        if not config.CONF.kubernetes.port_debug:
+            kuryr_ports = self._get_ports_by_attrs(
+                device_owner=kl_const.DEVICE_OWNER)
+            for port in kuryr_ports:
+                if port['id'] in self._recyclable_ports.keys():
+                    sg_current[port['id']] = port['security_groups']
+
+        for port_id, pool_key in self._recyclable_ports.copy().items():
+            if (not oslo_cfg.CONF.vif_pool.ports_pool_max or
+                self._get_pool_size(pool_key) <
+                    oslo_cfg.CONF.vif_pool.ports_pool_max):
+                port_name = (constants.KURYR_PORT_NAME
+                             if config.CONF.kubernetes.port_debug
+                             else '')
+                if (config.CONF.kubernetes.port_debug or
+                        list(pool_key[2]) != sg_current.get(port_id)):
+                    try:
+                        neutron.update_port(
+                            port_id,
+                            {
+                                "port": {
+                                    'name': port_name,
+                                    'device_id': '',
+                                    'security_groups': list(pool_key[2])
+                                }
+                            })
+                    except n_exc.NeutronClientException:
+                        LOG.warning("Error preparing port %s to be "
+                                    "reused, put back on the cleanable "
+                                    "pool.", port_id)
+                        continue
+                self._available_ports_pools.setdefault(
+                    pool_key, []).append(port_id)
+            else:
+                try:
+                    del self._existing_vifs[port_id]
+                    neutron.delete_port(port_id)
+                except n_exc.PortNotFoundClient:
+                    LOG.debug('Unable to release port %s as it no longer '
+                              'exists.', port_id)
+                except KeyError:
+                    LOG.debug('Port %s is not in the ports list.', port_id)
+            try:
+                del self._recyclable_ports[port_id]
+            except KeyError:
+                LOG.debug('Port already recycled: %s', port_id)
+
     def _recover_precreated_ports(self):
         if config.CONF.kubernetes.port_debug:
             available_ports = self._get_ports_by_attrs(
@@ -337,7 +377,7 @@ class NeutronVIFPool(BaseVIFPool):
         for port in available_ports:
             subnet_id = port['fixed_ips'][0]['subnet_id']
             subnet = {
-                subnet_id: default_subnet._get_subnet(subnet_id)}
+                subnet_id: utils.get_subnet(subnet_id)}
             vif_plugin = self._drv_vif._get_vif_plugin(port)
             vif = ovu.neutron_to_osvif_vif(vif_plugin, port, subnet)
             net_obj = subnet[subnet_id]
@@ -353,6 +393,29 @@ class NeutronVIFPool(BaseVIFPool):
         LOG.info("PORTS POOL: pools updated with pre-created ports")
         self._create_healthcheck_file()
 
+    def delete_network_pools(self, net_id):
+        neutron = clients.get_neutron_client()
+
+        # NOTE(ltomasbo): Note the pods should already be deleted, but their
+        # associated ports may not have been recycled yet, therefore not being
+        # on the available_ports_pools dict. The next call forces it to be on
+        # that dict before cleaning it up
+        self._trigger_return_to_pool()
+        for pool_key, ports_id in self._available_ports_pools.items():
+            if self._get_pool_key_net(pool_key) != net_id:
+                continue
+            self._available_ports_pools[pool_key] = []
+            for port_id in ports_id:
+                try:
+                    del self._existing_vifs[port_id]
+                except KeyError:
+                    LOG.debug('Port %s is not in the ports list.', port_id)
+                try:
+                    neutron.delete_port(port_id)
+                except n_exc.PortNotFoundClient:
+                    LOG.debug('Unable to release port %s as it no longer '
+                              'exists.', port_id)
+
 
 class NestedVIFPool(BaseVIFPool):
     """Manages VIFs for nested Kubernetes Pods.
@@ -385,7 +448,7 @@ class NestedVIFPool(BaseVIFPool):
                 port_id,
                 {
                     "port": {
-                        'name': pod['metadata']['name'],
+                        'name': c_utils.get_port_name(pod),
                     }
                 })
         # check if the pool needs to be populated
@@ -405,66 +468,77 @@ class NestedVIFPool(BaseVIFPool):
         If a maximun number of ports per pool is set, the port will be
         deleted if the maximun has been already reached.
         """
-        neutron = clients.get_neutron_client()
         while True:
-            sg_current = {}
-            if not config.CONF.kubernetes.port_debug:
-                kuryr_subports = self._get_ports_by_attrs(
-                    device_owner=['trunk:subport', kl_const.DEVICE_OWNER])
-                for subport in kuryr_subports:
-                    if subport['id'] in self._recyclable_ports.keys():
-                        sg_current[subport['id']] = subport['security_groups']
-
-            for port_id, pool_key in self._recyclable_ports.copy().items():
-                if (not oslo_cfg.CONF.vif_pool.ports_pool_max or
-                    self._get_pool_size(pool_key) <
-                        oslo_cfg.CONF.vif_pool.ports_pool_max):
-                    port_name = (constants.KURYR_PORT_NAME
-                                 if config.CONF.kubernetes.port_debug
-                                 else '')
-                    if (config.CONF.kubernetes.port_debug or
-                            list(pool_key[2]) != sg_current.get(port_id)):
-                        try:
-                            neutron.update_port(
-                                port_id,
-                                {
-                                    "port": {
-                                        'name': port_name,
-                                        'security_groups': list(pool_key[2])
-                                    }
-                                })
-                        except n_exc.NeutronClientException:
-                            LOG.warning("Error preparing port %s to be "
-                                        "reused, put back on the cleanable "
-                                        "pool.", port_id)
-                            continue
-                    self._available_ports_pools.setdefault(
-                        pool_key, []).append(port_id)
-                else:
-                    trunk_id = self._known_trunk_ids.get(pool_key, None)
-                    if not trunk_id:
-                        p_port = self._drv_vif._get_parent_port_by_host_ip(
-                            neutron, pool_key[0])
-                        trunk_id = self._drv_vif._get_trunk_id(p_port)
-                        self._known_trunk_ids[pool_key] = trunk_id
-                    try:
-                        self._drv_vif._remove_subport(neutron, trunk_id,
-                                                      port_id)
-                        self._drv_vif._release_vlan_id(
-                            self._existing_vifs[port_id].vlan_id)
-                        del self._existing_vifs[port_id]
-                        neutron.delete_port(port_id)
-                    except n_exc.PortNotFoundClient:
-                        LOG.debug('Unable to release port %s as it no longer '
-                                  'exists.', port_id)
-                    except KeyError:
-                        LOG.debug('Port %s is not in the ports list.', port_id)
-                    except n_exc.NeutronClientException:
-                        LOG.warning('Error removing the subport %s', port_id)
-                        continue
-                del self._recyclable_ports[port_id]
+            self._trigger_return_to_pool()
             eventlet.sleep(oslo_cfg.CONF.vif_pool.ports_pool_update_frequency)
 
+    @lockutils.synchronized('return_to_pool_nested')
+    def _trigger_return_to_pool(self):
+        neutron = clients.get_neutron_client()
+        sg_current = {}
+        if not config.CONF.kubernetes.port_debug:
+            kuryr_subports = self._get_ports_by_attrs(
+                device_owner=['trunk:subport', kl_const.DEVICE_OWNER])
+            for subport in kuryr_subports:
+                if subport['id'] in self._recyclable_ports.keys():
+                    sg_current[subport['id']] = subport['security_groups']
+
+        for port_id, pool_key in self._recyclable_ports.copy().items():
+            if (not oslo_cfg.CONF.vif_pool.ports_pool_max or
+                self._get_pool_size(pool_key) <
+                    oslo_cfg.CONF.vif_pool.ports_pool_max):
+                port_name = (constants.KURYR_PORT_NAME
+                             if config.CONF.kubernetes.port_debug
+                             else '')
+                if (config.CONF.kubernetes.port_debug or
+                        list(pool_key[2]) != sg_current.get(port_id)):
+                    try:
+                        neutron.update_port(
+                            port_id,
+                            {
+                                "port": {
+                                    'name': port_name,
+                                    'security_groups': list(pool_key[2])
+                                }
+                            })
+                    except n_exc.NeutronClientException:
+                        LOG.warning("Error preparing port %s to be "
+                                    "reused, put back on the cleanable "
+                                    "pool.", port_id)
+                        continue
+                self._available_ports_pools.setdefault(
+                    pool_key, []).append(port_id)
+            else:
+                trunk_id = self._get_trunk_id(neutron, pool_key)
+                try:
+                    self._drv_vif._remove_subport(neutron, trunk_id,
+                                                  port_id)
+                    self._drv_vif._release_vlan_id(
+                        self._existing_vifs[port_id].vlan_id)
+                    del self._existing_vifs[port_id]
+                    neutron.delete_port(port_id)
+                except n_exc.PortNotFoundClient:
+                    LOG.debug('Unable to release port %s as it no longer '
+                              'exists.', port_id)
+                except KeyError:
+                    LOG.debug('Port %s is not in the ports list.', port_id)
+                except n_exc.NeutronClientException:
+                    LOG.warning('Error removing the subport %s', port_id)
+                    continue
+            try:
+                del self._recyclable_ports[port_id]
+            except KeyError:
+                LOG.debug('Port already recycled: %s', port_id)
+
+    def _get_trunk_id(self, neutron, pool_key):
+        trunk_id = self._known_trunk_ids.get(pool_key, None)
+        if not trunk_id:
+            p_port = self._drv_vif._get_parent_port_by_host_ip(
+                neutron, pool_key[0])
+            trunk_id = self._drv_vif._get_trunk_id(p_port)
+            self._known_trunk_ids[pool_key] = trunk_id
+        return trunk_id
+
     def _get_parent_port_ip(self, port_id):
         neutron = clients.get_neutron_client()
         parent_port = neutron.show_port(port_id).get('port')
@@ -526,7 +600,7 @@ class NestedVIFPool(BaseVIFPool):
                     subnet_id = port['fixed_ips'][0]['subnet_id']
                     if not subnets.get(subnet_id):
                         subnets[subnet_id] = {subnet_id:
-                                              default_subnet._get_subnet(
+                                              utils.get_subnet(
                                                   subnet_id)}
         return parent_ports, subports, subnets
 
@@ -638,3 +712,106 @@ class NestedVIFPool(BaseVIFPool):
         no trunk is specified).
         """
         self._remove_precreated_ports(trunk_ips)
+
+    def delete_network_pools(self, net_id):
+        neutron = clients.get_neutron_client()
+        # NOTE(ltomasbo): Note the pods should already be deleted, but their
+        # associated ports may not have been recycled yet, therefore not being
+        # on the available_ports_pools dict. The next call forces it to be on
+        # that dict before cleaning it up
+        self._trigger_return_to_pool()
+        for pool_key, ports_ids in self._available_ports_pools.items():
+            if self._get_pool_key_net(pool_key) != net_id:
+                continue
+            self._available_ports_pools[pool_key] = []
+            trunk_id = self._get_trunk_id(neutron, pool_key)
+            try:
+                self._drv_vif._remove_subports(neutron, trunk_id, ports_ids)
+            except n_exc.NeutronClientException:
+                LOG.exception('Error removing subports from trunk: %s',
+                              trunk_id)
+                continue
+
+            for port_id in ports_ids:
+                try:
+                    self._drv_vif._release_vlan_id(
+                        self._existing_vifs[port_id].vlan_id)
+                    del self._existing_vifs[port_id]
+                except KeyError:
+                    LOG.debug('Port %s is not in the ports list.', port_id)
+                try:
+                    neutron.delete_port(port_id)
+                except n_exc.PortNotFoundClient:
+                    LOG.debug('Unable to delete subport %s as it no longer '
+                              'exists.', port_id)
+
+
+class MultiVIFPool(base.VIFPoolDriver):
+    """Manages pools with different VIF types.
+
+    It manages hybrid deployments containing both Bare Metal and Nested
+    Kubernetes Pods. To do that it creates a pool per node with a different
+    pool driver depending on the vif driver that the node is using.
+
+    It assumes a label pod_vif is added to each node to inform about the
+    driver set for that node. If no label is added, it assumes the default pod
+    vif: the one specified at kuryr.conf
+    """
+
+    def set_vif_driver(self):
+        self._vif_drvs = {}
+        pools_vif_drivers = oslo_cfg.CONF.vif_pool.pools_vif_drivers
+        if not pools_vif_drivers:
+            pod_vif = oslo_cfg.CONF.kubernetes.pod_vif_driver
+            drv_vif = base.PodVIFDriver.get_instance()
+            drv_pool = base.VIFPoolDriver.get_instance()
+            drv_pool.set_vif_driver(drv_vif)
+            self._vif_drvs[pod_vif] = drv_pool
+            return
+        for pool_driver, pod_driver in pools_vif_drivers.items():
+            if not utils.check_suitable_multi_pool_driver_opt(pool_driver,
+                                                              pod_driver):
+                LOG.error("The pool and pod driver selected are not "
+                          "compatible. They will be skipped")
+                raise exceptions.MultiPodDriverPoolConfigurationNotSupported()
+            drv_vif = base.PodVIFDriver.get_instance(driver_alias=pod_driver)
+            drv_pool = base.VIFPoolDriver.get_instance(
+                driver_alias=pool_driver)
+            drv_pool.set_vif_driver(drv_vif)
+            self._vif_drvs[pod_driver] = drv_pool
+
+    def request_vif(self, pod, project_id, subnets, security_groups):
+        pod_vif_type = self._get_pod_vif_type(pod)
+        return self._vif_drvs[pod_vif_type].request_vif(
+            pod, project_id, subnets, security_groups)
+
+    def release_vif(self, pod, vif, *argv):
+        pod_vif_type = self._get_pod_vif_type(pod)
+        self._vif_drvs[pod_vif_type].release_vif(pod, vif, *argv)
+
+    def activate_vif(self, pod, vif):
+        pod_vif_type = self._get_pod_vif_type(pod)
+        self._vif_drvs[pod_vif_type].activate_vif(pod, vif)
+
+    def delete_network_pools(self, net_id):
+        for vif_drv in self._vif_drvs.values():
+            if str(vif_drv) == 'NoopVIFPool':
+                continue
+            vif_drv.delete_network_pools(net_id)
+
+    def _get_pod_vif_type(self, pod):
+        node_name = pod['spec']['nodeName']
+        return self._get_node_vif_driver(node_name)
+
+    @MEMOIZE
+    def _get_node_vif_driver(self, node_name):
+        kubernetes = clients.get_kubernetes_client()
+        node_info = kubernetes.get(
+            constants.K8S_API_BASE + '/nodes/' + node_name)
+
+        labels = node_info['metadata'].get('labels', None)
+        if labels:
+            pod_vif = labels.get('pod_vif',
+                                 oslo_cfg.CONF.kubernetes.pod_vif_driver)
+            return pod_vif
+        return oslo_cfg.CONF.kubernetes.pod_vif_driver
