@@ -38,6 +38,7 @@ class LBaaSSpecHandler(k8s_base.ResourceEventHandler):
     """
 
     OBJECT_KIND = k_const.K8S_OBJ_SERVICE
+    OBJECT_WATCH_PATH = "%s/%s" % (k_const.K8S_API_BASE, "services")
 
     def __init__(self):
         super(LBaaSSpecHandler, self).__init__()
@@ -75,7 +76,7 @@ class LBaaSSpecHandler(k8s_base.ResourceEventHandler):
     def _has_selector(self, service):
         return service['spec'].get('selector')
 
-    def _has_clusterip(self, service):
+    def _has_clusterip(sef, service):
         # ignore headless service, clusterIP is None
         return service['spec'].get('clusterIP') != 'None'
 
@@ -220,6 +221,7 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
     """
 
     OBJECT_KIND = k_const.K8S_OBJ_ENDPOINTS
+    OBJECT_WATCH_PATH = "%s/%s" % (k_const.K8S_API_BASE, "endpoints")
 
     def __init__(self):
         super(LoadBalancerHandler, self).__init__()
@@ -273,7 +275,6 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
         # NOTE(ivc): deleting pool deletes its members
         if self._drv_lbaas.cascading_capable:
             self._drv_lbaas.release_loadbalancer(
-                endpoints=endpoints,
                 loadbalancer=lbaas_state.loadbalancer)
             if lbaas_state.service_pub_ip_info:
                 self._drv_service_pub_ip.release_pub_ip(
@@ -342,8 +343,7 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                                                              p.port]
             except KeyError:
                 continue
-        current_targets = {(str(m.ip), m.port, m.pool_id)
-                           for m in lbaas_state.members}
+        current_targets = {(str(m.ip), m.port) for m in lbaas_state.members}
 
         for subset in endpoints.get('subsets', []):
             subset_ports = subset.get('ports', [])
@@ -359,10 +359,10 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                     continue
                 for subset_port in subset_ports:
                     target_port = subset_port['port']
+                    if (target_ip, target_port) in current_targets:
+                        continue
                     port_name = subset_port.get('name')
                     pool = pool_by_tgt_name[port_name]
-                    if (target_ip, target_port, pool.id) in current_targets:
-                        continue
                     # TODO(apuimedo): Do not pass subnet_id at all when in
                     # L3 mode once old neutron-lbaasv2 is not supported, as
                     # octavia does not require it
@@ -376,13 +376,13 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                         # routed.
                         member_subnet_id = lbaas_state.loadbalancer.subnet_id
                     member = self._drv_lbaas.ensure_member(
-                        endpoints=endpoints,
                         loadbalancer=lbaas_state.loadbalancer,
                         pool=pool,
                         subnet_id=member_subnet_id,
                         ip=target_ip,
                         port=target_port,
-                        target_ref=target_ref)
+                        target_ref_namespace=target_ref['namespace'],
+                        target_ref_name=target_ref['name'])
                     lbaas_state.members.append(member)
                     changed = True
 
@@ -411,8 +411,7 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
         for member in lbaas_state.members:
             if (str(member.ip), member.port) in current_targets:
                 continue
-            self._drv_lbaas.release_member(endpoints,
-                                           lbaas_state.loadbalancer,
+            self._drv_lbaas.release_member(lbaas_state.loadbalancer,
                                            member)
             removed_ids.add(member.id)
         if removed_ids:
@@ -423,18 +422,18 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
     def _sync_lbaas_pools(self, endpoints, lbaas_state, lbaas_spec):
         changed = False
 
-        if self._remove_unused_pools(endpoints, lbaas_state, lbaas_spec):
+        if self._remove_unused_pools(lbaas_state, lbaas_spec):
             changed = True
 
         if self._sync_lbaas_listeners(endpoints, lbaas_state, lbaas_spec):
             changed = True
 
-        if self._add_new_pools(endpoints, lbaas_state, lbaas_spec):
+        if self._add_new_pools(lbaas_state, lbaas_spec):
             changed = True
 
         return changed
 
-    def _add_new_pools(self, endpoints, lbaas_state, lbaas_spec):
+    def _add_new_pools(self, lbaas_state, lbaas_spec):
         changed = False
 
         current_listeners_ids = {pool.listener_id
@@ -442,8 +441,7 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
         for listener in lbaas_state.listeners:
             if listener.id in current_listeners_ids:
                 continue
-            pool = self._drv_lbaas.ensure_pool(endpoints,
-                                               lbaas_state.loadbalancer,
+            pool = self._drv_lbaas.ensure_pool(lbaas_state.loadbalancer,
                                                listener)
             lbaas_state.pools.append(pool)
             changed = True
@@ -463,13 +461,12 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                     return True
         return False
 
-    def _remove_unused_pools(self, endpoints, lbaas_state, lbaas_spec):
+    def _remove_unused_pools(self, lbaas_state, lbaas_spec):
         removed_ids = set()
         for pool in lbaas_state.pools:
             if self._is_pool_in_spec(pool, lbaas_state, lbaas_spec):
                 continue
-            self._drv_lbaas.release_pool(endpoints,
-                                         lbaas_state.loadbalancer,
+            self._drv_lbaas.release_pool(lbaas_state.loadbalancer,
                                          pool)
             removed_ids.add(pool.id)
         if removed_ids:
@@ -504,10 +501,10 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                 continue
 
             listener = self._drv_lbaas.ensure_listener(
-                endpoints=endpoints,
                 loadbalancer=lbaas_state.loadbalancer,
                 protocol=protocol,
-                port=port)
+                port=port,
+                service_type=lbaas_spec.type)
             if listener is not None:
                 lbaas_state.listeners.append(listener)
                 changed = True
@@ -520,8 +517,7 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
         for listener in lbaas_state.listeners:
             if listener.id in current_listeners:
                 continue
-            self._drv_lbaas.release_listener(endpoints,
-                                             lbaas_state.loadbalancer,
+            self._drv_lbaas.release_listener(lbaas_state.loadbalancer,
                                              listener)
             removed_ids.add(listener.id)
         if removed_ids:
@@ -562,15 +558,17 @@ class LoadBalancerHandler(k8s_base.ResourceEventHandler):
                     lbaas_state.service_pub_ip_info)
 
             self._drv_lbaas.release_loadbalancer(
-                endpoints=endpoints,
                 loadbalancer=lb)
             lb = None
             changed = True
 
         if not lb:
             if lbaas_spec.ip:
+                lb_name = self._drv_lbaas.get_service_loadbalancer_name(
+                    endpoints['metadata']['namespace'],
+                    endpoints['metadata']['name'])
                 lb = self._drv_lbaas.ensure_loadbalancer(
-                    endpoints=endpoints,
+                    name=lb_name,
                     project_id=lbaas_spec.project_id,
                     subnet_id=lbaas_spec.subnet_id,
                     ip=lbaas_spec.ip,
