@@ -13,7 +13,6 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-from os_vif import objects as obj_vif
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
 
@@ -22,6 +21,9 @@ from kuryr_kubernetes import constants
 from kuryr_kubernetes.controller.drivers import base as drivers
 from kuryr_kubernetes import exceptions as k_exc
 from kuryr_kubernetes.handlers import k8s_base
+from kuryr_kubernetes import objects
+from kuryr_kubernetes import utils
+
 
 LOG = logging.getLogger(__name__)
 
@@ -33,24 +35,26 @@ class VIFHandler(k8s_base.ResourceEventHandler):
     the CNI driver (that runs on 'kubelet' nodes) is responsible for providing
     networking to Kubernetes pods. `VIFHandler` relies on a set of drivers
     (which are responsible for managing Neutron resources) to define the VIF
-    object and pass it to the CNI driver in form of the Kubernetes pod
+    objects and pass them to the CNI driver in form of the Kubernetes pod
     annotation.
     """
 
     OBJECT_KIND = constants.K8S_OBJ_POD
+    OBJECT_WATCH_PATH = "%s/%s" % (constants.K8S_API_BASE, "pods")
 
     def __init__(self):
         super(VIFHandler, self).__init__()
         self._drv_project = drivers.PodProjectDriver.get_instance()
         self._drv_subnets = drivers.PodSubnetsDriver.get_instance()
         self._drv_sg = drivers.PodSecurityGroupsDriver.get_instance()
-        self._drv_vif = drivers.PodVIFDriver.get_instance()
         # REVISIT(ltomasbo): The VIF Handler should not be aware of the pool
         # directly. Due to the lack of a mechanism to load and set the
         # VIFHandler driver, for now it is aware of the pool driver, but this
         # will be reverted as soon as a mechanism is in place.
-        self._drv_vif_pool = drivers.VIFPoolDriver.get_instance()
-        self._drv_vif_pool.set_vif_driver(self._drv_vif)
+        self._drv_vif_pool = drivers.VIFPoolDriver.get_instance(
+            driver_alias='multi_pool')
+        self._drv_vif_pool.set_vif_driver()
+        self._drv_multi_vif = drivers.MultiVIFDriver.get_enabled_drivers()
 
     def on_present(self, pod):
         if self._is_host_network(pod) or not self._is_pending_node(pod):
@@ -59,38 +63,64 @@ class VIFHandler(k8s_base.ResourceEventHandler):
             # where certain pods/namespaces/nodes can be managed by other
             # networking solutions/CNI drivers.
             return
+        state = self._get_pod_state(pod)
 
-        vif = self._get_vif(pod)
-
-        if not vif:
+        if not state:
             project_id = self._drv_project.get_project(pod)
             security_groups = self._drv_sg.get_security_groups(pod, project_id)
             subnets = self._drv_subnets.get_subnets(pod, project_id)
-            vif = self._drv_vif_pool.request_vif(pod, project_id, subnets,
-                                                 security_groups)
+
+            # Request the default interface of pod
+            main_vif = self._drv_vif_pool.request_vif(
+                pod, project_id, subnets, security_groups)
+
+            state = objects.vif.PodState(default_vif=main_vif)
+
+            # Request the additional interfaces from multiple dirvers
+            additional_vifs = []
+            for driver in self._drv_multi_vif:
+                additional_vifs.extend(
+                    driver.request_additional_vifs(
+                        pod, project_id, security_groups))
+            if additional_vifs:
+                state.additional_vifs = {}
+                for i, vif in enumerate(additional_vifs, start=1):
+                    k = constants.ADDITIONAL_IFNAME_PREFIX + str(i)
+                    state.additional_vifs[k] = vif
+
             try:
-                self._set_vif(pod, vif)
+                self._set_pod_state(pod, state)
             except k_exc.K8sClientException as ex:
                 LOG.debug("Failed to set annotation: %s", ex)
                 # FIXME(ivc): improve granularity of K8sClient exceptions:
                 # only resourceVersion conflict should be ignored
-                self._drv_vif_pool.release_vif(pod, vif, project_id,
-                                               security_groups)
-        elif not vif.active:
-            self._drv_vif_pool.activate_vif(pod, vif)
-            self._set_vif(pod, vif)
+                for ifname, vif in state.vifs.items():
+                    self._drv_for_vif(vif).release_vif(pod, vif, project_id,
+                                                       security_groups)
+        else:
+            changed = False
+            for ifname, vif in state.vifs.items():
+                if not vif.active:
+                    self._drv_for_vif(vif).activate_vif(pod, vif)
+                    changed = True
+            if changed:
+                self._set_pod_state(pod, state)
 
     def on_deleted(self, pod):
         if self._is_host_network(pod):
             return
+        project_id = self._drv_project.get_project(pod)
+        security_groups = self._drv_sg.get_security_groups(pod, project_id)
 
-        vif = self._get_vif(pod)
+        state = self._get_pod_state(pod)
+        if state:
+            for ifname, vif in state.vifs.items():
+                self._drv_for_vif(vif).release_vif(pod, vif, project_id,
+                                                   security_groups)
 
-        if vif:
-            project_id = self._drv_project.get_project(pod)
-            security_groups = self._drv_sg.get_security_groups(pod, project_id)
-            self._drv_vif_pool.release_vif(pod, vif, project_id,
-                                           security_groups)
+    def _drv_for_vif(self, vif):
+        # TODO(danil): a better polymorphism is required here
+        return self._drv_vif_pool
 
     @staticmethod
     def _is_host_network(pod):
@@ -105,29 +135,35 @@ class VIFHandler(k8s_base.ResourceEventHandler):
         except KeyError:
             return False
 
-    def _set_vif(self, pod, vif):
+    def _set_pod_state(self, pod, state):
         # TODO(ivc): extract annotation interactions
-        if vif is None:
-            LOG.debug("Removing VIF annotation: %r", vif)
+        if not state:
+            LOG.debug("Removing VIFs annotation: %r", state)
             annotation = None
         else:
-            vif.obj_reset_changes(recursive=True)
-            LOG.debug("Setting VIF annotation: %r", vif)
-            annotation = jsonutils.dumps(vif.obj_to_primitive(),
-                                         sort_keys=True)
+            state_dict = state.obj_to_primitive()
+            annotation = jsonutils.dumps(state_dict, sort_keys=True)
+            LOG.debug("Setting VIFs annotation: %r", annotation)
+
+        # NOTE(dulek): We don't care about compatiblity with Queens format
+        #              here, as eventually all Kuryr services will be upgraded
+        #              and cluster will start working normally. Meanwhile
+        #              we just ignore issue of old services being unable to
+        #              read new annotations.
+
         k8s = clients.get_kubernetes_client()
         k8s.annotate(pod['metadata']['selfLink'],
                      {constants.K8S_ANNOTATION_VIF: annotation},
                      resource_version=pod['metadata']['resourceVersion'])
 
-    def _get_vif(self, pod):
+    def _get_pod_state(self, pod):
         # TODO(ivc): same as '_set_vif'
         try:
             annotations = pod['metadata']['annotations']
-            vif_annotation = annotations[constants.K8S_ANNOTATION_VIF]
+            state_annotation = annotations[constants.K8S_ANNOTATION_VIF]
         except KeyError:
             return None
-        vif_dict = jsonutils.loads(vif_annotation)
-        vif = obj_vif.vif.VIFBase.obj_from_primitive(vif_dict)
-        LOG.debug("Got VIF from annotation: %r", vif)
-        return vif
+        state_annotation = jsonutils.loads(state_annotation)
+        state = utils.extract_pod_annotation(state_annotation)
+        LOG.debug("Got VIFs from annotation: %r", state)
+        return state
