@@ -13,6 +13,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import sys
 import time
 
 from kuryr_kubernetes import clients
@@ -55,7 +56,8 @@ class Watcher(health.HealthHandler):
     graceful=False)` for asynchronous `Watcher`).
     """
 
-    def __init__(self, handler, thread_group=None, timeout=None):
+    def __init__(self, handler, thread_group=None, timeout=None,
+                 exit_on_stop=False):
         """Initializes a new Watcher instance.
 
         :param handler: a `callable` object to be invoked for each observed
@@ -82,6 +84,7 @@ class Watcher(health.HealthHandler):
         if timeout is None:
             timeout = CONF.kubernetes.watch_retry_timeout
         self._timeout = timeout
+        self._exit_on_stop = exit_on_stop
 
     def add(self, path):
         """Adds ths K8s resource to the Watcher.
@@ -108,6 +111,9 @@ class Watcher(health.HealthHandler):
         if path in self._watching:
             self._stop_watch(path)
 
+    def is_running(self):
+        return self._running
+
     def start(self):
         """Starts the Watcher.
 
@@ -143,11 +149,20 @@ class Watcher(health.HealthHandler):
 
     def _graceful_watch_exit(self, path):
         try:
+            self.remove(path)
             self._watching.pop(path)
             self._idle.pop(path)
             LOG.info("Stopped watching '%s'", path)
         except KeyError:
             LOG.error("Failed to exit watch gracefully")
+        finally:
+            if not self._watching and not self._idle:
+                self.stop()
+                if self._exit_on_stop:
+                    LOG.info("No remaining active watchers, Exiting...")
+                    # TODO(dulek): This complicates things, remove once we
+                    #              don't support running without kuryr-daemon.
+                    sys.exit(1)
 
     def _watch(self, path):
         attempts = 0
@@ -176,11 +191,13 @@ class Watcher(health.HealthHandler):
                     self._idle[path] = True
                     if not (self._running and path in self._resources):
                         return
-            except Exception as e:
-                LOG.warning("Restarting(%s) watching '%s': %s",
-                            attempts, path, e)
+            except Exception:
+                LOG.exception("Caught exception while watching.")
+                LOG.warning("Restarting(%s) watching '%s'.",
+                            attempts, path)
                 attempts += 1
                 retry = True
+                self._idle[path] = True
             finally:
                 if not retry:
                     self._graceful_watch_exit(path)
